#!/usr/bin/python3

import gettext

from gi.repository import Gio

import config
import misc
import util
from util import OpStatus
import prefs

_ = gettext.gettext

class NewOpUserNotification():
    def __init__(self, op):
        self.op = op

        self.send_notification()

    @misc._idle
    def send_notification(self):
        if prefs.get_show_notifications():
            notification = Gio.Notification.new(_("New incoming files"))

            if prefs.require_permission_for_transfer():
                # John Johnson would like to send you 1 file (foobar.txt)
                # John Johnson would like to send you 5 files
                body =gettext.ngettext(
                    _("%(sender_name)s would like to send you %(file_count)d file (%(file_name)s)."),
                    _("%(sender_name)s would like to send you %(file_count)d files"),self.op.total_count) \
                        % ({
                              "sender_name": self.op.sender_name,
                              "file_count": self.op.total_count,
                              "file_name": self.op.top_dir_basenames[0]
                          })

                notification.set_body(body)
                notification.set_icon(Gio.ThemedIcon(name="org.x.Warpinator-symbolic"))

                notification.add_button(_("Accept"), "app.notification-response::accept")
                notification.add_button(_("Decline"), "app.notification-response::decline")
                notification.set_default_action("app.notification-response::focus")

                notification.set_priority(Gio.NotificationPriority.URGENT)

                app = Gio.Application.get_default()
                app.lookup_action("notification-response").connect("activate", self._notification_response, self.op)

                self.op.connect("status-changed", \
                   lambda op: app.withdraw_notification(op.sender) if self.op.status != OpStatus.WAITING_PERMISSION else None)
            else:
                # John Johnson is sending you 1 file (foobar.txt)
                # John Johnson is sending you 5 files
                body =gettext.ngettext(
                    _("%(sender_name)s is sending you %(file_count)d file (%(file_name)s)."),
                    _("%(sender_name)s is sending you %(file_count)d files"), self.op.total_count) \
                        % ({
                              "sender_name": self.op.sender_name,
                              "file_count": self.op.total_count,
                              "file_name": self.op.top_dir_basenames[0]
                          })

                notification.set_body(body)
                notification.set_icon(Gio.ThemedIcon(name="org.x.Warpinator-symbolic"))

            Gio.Application.get_default().send_notification(self.op.sender, notification)

    def _notification_response(self, action, variant, op):
        response = variant.unpack()

        if response == "accept":
            op.accept_transfer()
        elif response == "decline":
            op.decline_transfer_request()
        else:
            op.focus()

        app = Gio.Application.get_default()
        app.lookup_action("notification-response").disconnect_by_func(self._notification_response)

class TransferCompleteNotification():
    def __init__(self, op, sender=True, warn=False):
        self.op = op
        self.sender = sender
        self.warn = warn

        self.send_notification()

    @misc._idle
    def send_notification(self):
        if prefs.get_show_notifications():
            notification = Gio.Notification.new(_("Transfer complete"))
            if self.sender:
                if self.warn:
                    body = (_("The transfer to %s has finished, but with errors") % self.op.receiver_name)
                else:
                    body = (_("The transfer to %s has finished successfully") % self.op.receiver_name)
            else:
                if self.warn:
                    body = (_("The transfer from %s has finished, but with errors") % self.op.sender_name)
                else:
                    body = (_("The transfer from %s has finished successfully") % self.op.sender_name)

            notification.set_body(body)

            if self.warn:
                icon_name = "dialog-warning-symbolic"
            else:
                icon_name = "emblem-ok-symbolic"
            notification.set_icon(Gio.ThemedIcon(name=icon_name))
            notification.set_default_action("app.notification-response::focus")

            notification.set_priority(Gio.NotificationPriority.NORMAL)

            app = Gio.Application.get_default()
            app.lookup_action("notification-response").connect("activate", self._notification_response, self.op)

            app.get_default().send_notification(self.op.sender, notification)

    def _notification_response(self, action, variant, op):
        op.focus()

        app = Gio.Application.get_default()
        app.lookup_action("notification-response").disconnect_by_func(self._notification_response)

class TransferFailedNotification():
    def __init__(self, op, sender=True):
        self.op = op
        self.sender = sender

        self.send_notification()

    @misc._idle
    def send_notification(self):
        if prefs.get_show_notifications():
            notification = Gio.Notification.new(_("Transfer failed"))

            if self.sender:
                body = (_("Something went wrong with the transfer to %s") % self.op.receiver_name)
            else:
                body = (_("Something went wrong with the transfer from %s") % self.op.sender_name)

            notification.set_body(body)
            notification.set_icon(Gio.ThemedIcon(name="dialog-error-symbolic"))
            notification.set_default_action("app.notification-response::focus")

            notification.set_priority(Gio.NotificationPriority.NORMAL)

            app = Gio.Application.get_default()
            app.lookup_action("notification-response").connect("activate", self._notification_response, self.op)

            app.get_default().send_notification(self.op.sender, notification)

    def _notification_response(self, action, variant, op):
        op.focus()

        app = Gio.Application.get_default()
        app.lookup_action("notification-response").disconnect_by_func(self._notification_response)

class TransferStoppedNotification():
    def __init__(self, op, sender=True):
        self.op = op
        self.sender = sender

        self.send_notification()

    @misc._idle
    def send_notification(self):
        if prefs.get_show_notifications():
            notification = Gio.Notification.new(_("Transfer cancelled"))

            if self.sender:
                body = (_("Your transfer to %s was cancelled") % self.op.receiver_name)
            else:
                body = (_("An incoming transfer from %s was cancelled") % self.op.sender_name)

            notification.set_body(body)
            notification.set_icon(Gio.ThemedIcon(name="dialog-info-symbolic"))
            notification.set_default_action("app.notification-response::focus")

            notification.set_priority(Gio.NotificationPriority.NORMAL)

            app = Gio.Application.get_default()
            app.lookup_action("notification-response").connect("activate", self._notification_response, self.op)

            app.get_default().send_notification(self.op.sender, notification)

    def _notification_response(self, action, variant, op):
        op.focus()

        app = Gio.Application.get_default()
        app.lookup_action("notification-response").disconnect_by_func(self._notification_response)

class WarpinatorSendNotification():
    def __init__(self, op):
        self.op = op

        self.send_notification()

    @misc._idle
    def send_notification(self):
        if prefs.get_show_notifications():
            notification = Gio.Notification.new(_("Sending files"))

            if self.op.total_count > 1:
                body = (_("Sending %d files to %s") % (self.op.total_count, self.op.receiver_name))
            else:
                body = (_("Sending '%s' to %s") % (self.op.top_dir_basenames[0], self.op.sender_name))

            notification.set_body(body)
            notification.set_icon(Gio.ThemedIcon(name="dialog-info-symbolic"))
            notification.set_default_action("app.notification-response::focus")

            notification.set_priority(Gio.NotificationPriority.NORMAL)

            app = Gio.Application.get_default()
            app.lookup_action("notification-response").connect("activate", self._notification_response, self.op)

            app.get_default().send_notification(self.op.sender, notification)

    def _notification_response(self, action, variant, op):
        op.focus()

        app = Gio.Application.get_default()
        app.lookup_action("notification-response").disconnect_by_func(self._notification_response)



#!/usr/bin/python3

import os
import threading
import gettext
import logging
import time
import pkg_resources
from concurrent import futures

from gi.repository import GObject, GLib

import grpc
import warp_pb2
import warp_pb2_grpc

import config
import auth
import interceptors
import networkmonitor
import remote
import remote_registration
import prefs
import util
import misc
import transfers
from ops import ReceiveOp
from util import TransferDirection, OpStatus, RemoteStatus

try:
    import zeroconf_
    from zeroconf_ import ServiceInfo, Zeroconf, ServiceBrowser
except:
    import zeroconf

    zc_version = pkg_resources.parse_version(zeroconf.__version__)
    zc_min_version = pkg_resources.parse_version("0.27.0")

    if zc_version < zc_min_version:
        print("Python3 zeroconf must be >= %s" % zc_min_version)
        exit(1)

    from zeroconf import ServiceInfo, Zeroconf, ServiceBrowser

_ = gettext.gettext

void = warp_pb2.VoidType()

SERVICE_TYPE = "_warpinator._tcp.local."

# server (this is on a separate thread from the ui, grpc isn't compatible with
# gmainloop)
class Server(threading.Thread, warp_pb2_grpc.WarpServicer, GObject.Object):
    __gsignals__ = {
        "remote-machine-added": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "remote-machine-removed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "remote-machine-ops-changed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "local-info-changed": (GObject.SignalFlags.RUN_LAST, None, (str,)),
        "server-started": (GObject.SignalFlags.RUN_LAST, None, ()),
        "shutdown-complete": (GObject.SignalFlags.RUN_LAST, None, ())
    }
    def __init__(self, ip_info, port, auth_port):
        threading.Thread.__init__(self, name="server-thread")
        super(Server, self).__init__()
        GObject.Object.__init__(self)

        self.service_name = None
        self.service_ident = None

        self.ip_info = ip_info
        self.port = port
        self.auth_port = auth_port

        self.untrusted_remote_machines = {}
        self.remote_machines = {}
        self.remote_registrar = None

        self.server_thread_keepalive = threading.Event()

        self.netmon = networkmonitor.get_network_monitor()

        self.server = None
        self.browser = None
        self.zeroconf = None
        self.info = None

        self.display_name = GLib.get_real_name()
        self.start()

    def start_zeroconf(self):
        try:
            logging.info("Using bundled zeroconf v%s" % zeroconf_.__version__)
        except:
            logging.info("Using system zeroconf v%s" % zeroconf.__version__)

        self.zeroconf = Zeroconf(interfaces=[self.ip_info.ip4_address])

        self.service_ident = prefs.get_connect_id()
        self.service_name = "%s.%s" % (self.service_ident, SERVICE_TYPE)

        # If this process is killed (either kill or network issue), the service
        # never gets unregistered, which will prevent remotes from seeing us
        # when we come back.  Our first service info is to get us back on
        # momentarily, and the unregister properly, so remotes get notified.
        # Then we'll do it again without the flush property for the real
        # connection.

        init_info = ServiceInfo(SERVICE_TYPE,
                                self.service_name,
                                port=self.port,
                                addresses=self.ip_info.as_binary_list(),
                                properties={ 'hostname': util.get_hostname(),
                                             'type': 'flush' })

        self.zeroconf.register_service(init_info)
        time.sleep(3)
        self.zeroconf.unregister_service(init_info)
        time.sleep(3)

        self.info = ServiceInfo(SERVICE_TYPE,
                                self.service_name,
                                port=self.port,
                                addresses=self.ip_info.as_binary_list(),
                                properties={ 'hostname': util.get_hostname(),
                                             'api-version': config.RPC_API_VERSION,
                                             'auth-port': str(prefs.get_auth_port()),
                                             'type': 'real' })

        self.zeroconf.register_service(self.info)
        self.browser = ServiceBrowser(self.zeroconf, SERVICE_TYPE, self, addr=self.ip_info.ip4_address)

        return False

    # Will be mandatory eventually, this will have to be here even if we don't care about it.
    def update_service(self, zeroconf, _type, name):
        pass

    # Zeroconf worker thread
    def remove_service(self, zeroconf, _type, name):
        if name == self.service_name:
            return

        ident = name.partition(".%s" % SERVICE_TYPE)[0]

        try:
            remote = self.remote_machines[ident]
        except KeyError:
            logging.debug(">>> Discovery: unknown service ident (%s) reported as gone by zc." % ident)
            return

        logging.debug(">>> Discovery: service %s (%s:%d) has disappeared."
                          % (remote.display_hostname, remote.ip_info.ip4_address, remote.port))

        remote.has_zc_presence = False

    # Zeroconf worker thread
    def add_service(self, zeroconf, _type, name):
        info = zeroconf.get_service_info(_type, name)

        if info:
            ident = name.partition(".%s" % SERVICE_TYPE)[0]

            try:
                remote_hostname = info.properties[b"hostname"].decode()
            except KeyError:
                logging.critical(">>> Discovery: no hostname in service info properties.  Is this an old version?")
                return

            remote_ip_info = util.RemoteInterfaceInfo(info.addresses)

            if remote_ip_info == self.ip_info:
                return

            try:
                # Check if this is a flush registration to reset the remote server's presence.
                if info.properties[b"type"].decode() == "flush":
                    logging.debug(">>> Discovery: received flush service info (ignoring): %s (%s:%d)"
                                      % (remote_hostname, remote_ip_info.ip4_address, info.port))
                    return
            except KeyError:
                logging.warning("No type in service info properties, assuming this is a real connect attempt")

            if ident == self.service_ident:
                return

            try:
                api_version = info.properties[b"api-version"].decode()
                auth_port = int(info.properties[b"auth-port"].decode())
            except KeyError:
                api_version = "1"
                auth_port = 0

            # FIXME: I'm not sure why we still get discovered by other networks in some cases -
            # The Zeroconf object has a specific ip it is set to, what more do I need to do?
            if not self.netmon.same_subnet(remote_ip_info):
                logging.debug(">>> Discovery: service is not on this subnet, ignoring: %s (%s)" % (remote_hostname, remote_ip_info.ip4_address))
                return

            try:
                machine = self.remote_machines[ident]
                machine.has_zc_presence = True
                logging.info(">>> Discovery: existing remote: %s (%s:%d)"
                                  % (machine.display_hostname, remote_ip_info.ip4_address, info.port))

                # If the remote truly is the same one (our service info just dropped out
                # momentarily), this will end up just retrieving the current cert again.
                # If this was a real disconnect we didn't notice, we'll have the new cert
                # which we'll need when our supposedly existing connection tries to continue
                # pinging. It will fail out and restart the connection loop, and will need
                # this updated one.

                # This blocks the zeroconf thread.
                if not self.remote_registrar.register(ident, remote_hostname, remote_ip_info, info.port, auth_port, api_version) or self.server_thread_keepalive.is_set():
                    logging.warning("Register failed, or the server was shutting down during registration, ignoring remote %s (%s:%d) auth port: %d"
                                     % (remote_hostname, remote_ip_info.ip4_address, info.port, auth_port))
                    return

                if machine.status == RemoteStatus.ONLINE:
                    logging.debug(">>> Discovery: rejoining existing connect with %s (%s:%d)"
                                  % (machine.display_hostname, remote_ip_info.ip4_address, info.port))
                    return

                # Update our connect info if it changed.
                machine.hostname = remote_hostname
                machine.ip_info = remote_ip_info
                machine.port = info.port
                machine.api_version = api_version
            except KeyError:
                display_hostname = remote_hostname
                i = 1

                while True:
                    found = False

                    for key in self.remote_machines.keys():
                        remote_machine = self.remote_machines[key]

                        if remote_machine.display_hostname == display_hostname:
                            display_hostname = "%s[%d]" % (remote_hostname, i)
                            found = True
                            break

                    i += 1

                    if not found:
                        break

                logging.info(">>> Discovery: new remote: %s (%s:%d)"
                                  % (display_hostname, remote_ip_info.ip4_address, info.port))

                machine = remote.RemoteMachine(ident,
                                               remote_hostname,
                                               display_hostname,
                                               remote_ip_info,
                                               info.port,
                                               self.service_ident,
                                               api_version)

                # This blocks the zeroconf thread. Registration will timeout
                if not self.remote_registrar.register(ident, remote_hostname, remote_ip_info, info.port, auth_port, api_version) or self.server_thread_keepalive.is_set():
                    logging.debug("Register failed, or the server was shutting down during registration, ignoring remote %s (%s:%d) auth port: %d"
                                     % (remote_hostname, remote_ip_info.ip4_address, info.port, auth_port))
                    return

                self.remote_machines[ident] = machine
                machine.connect("ops-changed", self.remote_ops_changed)
                machine.connect("remote-status-changed", self.remote_status_changed)
                self.idle_emit("remote-machine-added", machine)

            machine.has_zc_presence = True

            machine.shutdown() # This does nothing if run more than once.  It's here to make sure
                               # the previous start thread is complete before starting a new one.
                               # This is needed in the corner case where the remote has gone offline,
                               # and returns before our Ping loop times out and closes the thread
                               # itself.

            machine.start_remote_thread()

    def run(self):
        logging.debug("Server: starting server on %s (%s)" % (self.ip_info.ip4_address, self.ip_info.iface))
        logging.info("Using api version %s" % config.RPC_API_VERSION)
        logging.info("Our uuid: %s" % prefs.get_connect_id())


        self.remote_registrar = remote_registration.Registrar(self.ip_info, self.port, self.auth_port)
        util.initialize_rpc_threadpool()

        options=(
            ('grpc.keepalive_time_ms', 10 * 1000),
            ('grpc.keepalive_timeout_ms', 5 * 1000),
            ('grpc.keepalive_permit_without_calls', True),
            ('grpc.http2.max_pings_without_data', 0),
            ('grpc.http2.min_time_between_pings_ms', 10 * 1000),
            ('grpc.http2.min_ping_interval_without_data_ms',  5 * 1000)
        )

        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=prefs.get_server_pool_max_threads()),
                                  options=options,
                                  interceptors=[interceptors.ChunkCompressor()])
        warp_pb2_grpc.add_WarpServicer_to_server(self, self.server)

        pair = auth.get_singleton().get_server_creds()
        server_credentials = grpc.ssl_server_credentials((pair,))

        if self.ip_info.ip4_address:
            self.server.add_secure_port('%s:%d' % (self.ip_info.ip4_address, self.port),
                                        server_credentials)
        # if self.ip_info.ip6_address:
        #     self.server.add_secure_port('%s:%d' % (self.ip_info.ip6_address, self.port),
        #                                 server_credentials)
        self.server.start()

        self.server_thread_keepalive.clear()

        try:
            self.start_zeroconf()
        except Exception as e:
            logging.critical("Zeroconf failed to start, server will terminate: %s" % e)
            self.server_thread_keepalive.set()

        self.idle_emit("server-started")

        logging.info("Server: ACTIVE")

        # **** RUNNING ****
        while not self.server_thread_keepalive.is_set():
            self.server_thread_keepalive.wait(10)
        # **** STOPPING ****

        self.remote_registrar.shutdown_registration_servers()
        self.remote_registrar = None

        logging.debug("Server: stopping discovery and advertisement")

        # If the network is down, this will probably print an exception - it's ok,
        # zeroconf catches it.
        try:
            self.zeroconf.close()
        except:
            logging.critical("Can't close Zeroconf - maybe it failed to start")
            pass

        remote_machines = list(self.remote_machines.values())
        for remote in remote_machines:
            self.idle_emit("remote-machine-removed", remote)
            logging.debug("Server: Closing connection to remote machine %s (%s:%d)"
                              % (remote.display_hostname, remote.ip_info.ip4_address, remote.port))

            remote.shutdown()

        remote_machines = None

        logging.debug("Server: terminating server")
        self.server.stop(grace=2).wait()

        self.idle_emit("shutdown-complete")
        self.server = None

        util.global_rpc_threadpool.shutdown(wait=True)
        logging.debug("Server: server stopped")

    def shutdown(self):
        self.server_thread_keepalive.set()

    def remote_status_changed(self, remote):
        if remote.status == RemoteStatus.OFFLINE:
            self.emit("remote-machine-removed", remote)

    def add_receive_op_to_remote_machine(self, op):
        self.remote_machines[op.sender].add_op(op)

    @misc._idle
    def remote_ops_changed(self, remote_machine):
        self.emit("remote-machine-ops-changed", remote_machine.ident)

    def list_remote_machines(self):
        return self.remote_machines.values()

    def get_active_op_count(self, incoming_only=False):
        count = 0

        for machine in self.remote_machines.values():
            if machine.status != RemoteStatus.ONLINE:
                continue
            for op in machine.transfer_ops:
                if incoming_only and not isinstance(op, ReceiveOp):
                    continue
                if op.status == OpStatus.TRANSFERRING:
                    count += 1

        return count

    def cancel_all_ops(self):
        for machine in self.remote_machines.values():
            if machine.status != RemoteStatus.ONLINE:
                continue
            for op in machine.transfer_ops:
                if op.status == OpStatus.TRANSFERRING:
                    op.stop_transfer()

    @misc._idle
    def idle_emit(self, signal, *callback_data):
        self.emit(signal, *callback_data)

    def Ping(self, request, context):
        logging.debug("Server Ping: from %s" % request.readable_name)

        try:
            remote = self.remote_machines[request.id]
        except KeyError as e:
            logging.debug("Server Ping: ping is from unknown remote (or not fully online yet)")

        return void

    def CheckDuplexConnection(self, request, context):
        logging.debug("Server RPC: CheckDuplexConnection from '%s'" % request.readable_name)
        response = False

        try:
            remote = self.remote_machines[request.id]
            response = (remote.status in (RemoteStatus.AWAITING_DUPLEX, RemoteStatus.ONLINE))
        except KeyError:
            pass

        return warp_pb2.HaveDuplex(response=response)

    def WaitingForDuplex(self, request, context):
        logging.debug("Server RPC: WaitingForDuplex from '%s' (api v2)" % request.readable_name)

        max_tries = 20
        i = 0

        # try for ~5 seconds (the caller aborts at 4)
        while i < max_tries:
            response = False

            try:
                remote = self.remote_machines[request.id]
                response = (remote.status in (RemoteStatus.AWAITING_DUPLEX, RemoteStatus.ONLINE))
            except KeyError:
                pass

            if response:
                break
            else:
                i += 1
                if i == max_tries:
                    context.abort(code=grpc.StatusCode.DEADLINE_EXCEEDED,
                                  details='Server timed out while waiting for his corresponding remote to connect back to you.')
                    return
                time.sleep(.25)

        return warp_pb2.HaveDuplex(response=response)

    def GetRemoteMachineInfo(self, request, context):
        logging.debug("Server RPC: GetRemoteMachineInfo from '%s'" % request.readable_name)

        return warp_pb2.RemoteMachineInfo(display_name=GLib.get_real_name(),
                                          user_name=GLib.get_user_name())

    def GetRemoteMachineAvatar(self, request, context):
        logging.debug("Server RPC: GetRemoteMachineAvatar from '%s'" % request.readable_name)

        path = os.path.join(GLib.get_home_dir(), ".face")
        if os.path.exists(path):
            return transfers.load_file_in_chunks(path)
        else:
            context.abort(code=grpc.StatusCode.NOT_FOUND, details='.face file not found!')

    def ProcessTransferOpRequest(self, request, context):
        logging.debug("Server RPC: ProcessTransferOpRequest from '%s'" % request.info.readable_name)

        remote_machine = self.remote_machines[request.info.ident]

        try:
            remote_machine = self.remote_machines[request.info.ident]
        except KeyError as e:
            logging.warning("Received transfer op request for unknown remote: %s" % e)
            return

        for existing_op in remote_machine.transfer_ops:
            if existing_op.start_time == request.info.timestamp:
                # Compression could have changed for a restart, as it's not tied to the op.
                try:
                    existing_op.use_compression = request.info.use_compression
                except AttributeError:
                    existing_op.use_compression = False
                existing_op.set_status(OpStatus.WAITING_PERMISSION)
                self.add_receive_op_to_remote_machine(existing_op)
                return void

        op = ReceiveOp(request.info.ident)

        op.start_time = request.info.timestamp

        op.sender_name = request.sender_name
        op.receiver = request.receiver
        op.receiver_name = request.receiver_name
        op.status = OpStatus.WAITING_PERMISSION
        op.total_size = request.size
        op.total_count = op.remaining_count = request.count
        op.mime_if_single = request.mime_if_single
        op.name_if_single = request.name_if_single
        op.top_dir_basenames = request.top_dir_basenames

        # If request.info (grpc_pb2.OpInfo) doesn't have a use_compression field,
        # If doesn't support compression (older version of warp).
        try:
            op.use_compression = request.info.use_compression
        except AttributeError:
            op.use_compression = False

        op.connect("initial-setup-complete", self.add_receive_op_to_remote_machine)
        op.prepare_receive_info()

        return void

    def CancelTransferOpRequest(self, request, context):### good
        logging.debug("Server RPC: CancelTransferOpRequest from '%s'" % request.readable_name)

        try:
            op = self.remote_machines[request.ident].lookup_op(request.timestamp)
        except KeyError as e:
            logging.warning("Received cancel transfer op request for unknown op: %s" % e)
            return

        # If we receive this call, this means the op was cancelled remotely.  So,
        # our op with TO_REMOTE_MACHINE (we initiated it) was cancelled by the recipient.
        if op.direction == TransferDirection.TO_REMOTE_MACHINE:
            op.set_status(OpStatus.CANCELLED_PERMISSION_BY_RECEIVER)
        else:
            op.set_status(OpStatus.CANCELLED_PERMISSION_BY_SENDER)

        return void

    # receiver server responders
    def StartTransfer(self, request, context):
        logging.debug("Server RPC: StartTransfer from '%s'" % request.readable_name)

        start_time = GLib.get_monotonic_time()

        try:
            remote = self.remote_machines[request.ident]
        except KeyError as e:
            logging.warning("Server: start transfer is from unknown remote: %s" % e)
            return

        try:
            op = self.remote_machines[request.ident].lookup_op(request.timestamp)
        except KeyError as e:
            logging.warning("Server: start transfer for unknowns op: %s" % e)
            return

        cancellable = threading.Event()
        op.file_send_cancellable = cancellable

        op.set_status(OpStatus.TRANSFERRING)

        op.progress_tracker = transfers.OpProgressTracker(op)
        op.current_progress_report = None
        sender = transfers.FileSender(op, request.timestamp, cancellable)

        def transfer_done():
            if sender.error is not None:
                op.set_error(sender.error)
                op.set_status(OpStatus.FAILED_UNRECOVERABLE)
            elif op.file_send_cancellable.is_set():
                logging.debug("Server: file send cancelled")
            else:
                logging.debug("Server: transfer of %s files (%s) finished in %s" % \
                    (op.total_count, GLib.format_size(op.total_size),\
                     util.precise_format_time_span(GLib.get_monotonic_time() - start_time)))

        context.add_callback(transfer_done)
        return sender.read_chunks()

    def StopTransfer(self, request, context):
        logging.debug("Server RPC: StopTransfer from '%s'" % request.info.readable_name)

        try:
            op = self.remote_machines[request.info.ident].lookup_op(request.info.timestamp)
        except KeyError as e:
            logging.warning("Server: stop transfer was for unknown op: %s" % e)
            return

        # If we receive this call, this means the op was stopped remotely.  So,
        # our op with TO_REMOTE_MACHINE (we initiated it) was cancelled by the recipient.

        if request.error:
            op.error_msg = _("An error occurred on the remote machine")

        if op.direction == TransferDirection.TO_REMOTE_MACHINE:
            if op.file_send_cancellable is not None:
                op.file_send_cancellable.set()
            logging.debug("Server: sender received stop transfer by receiver: %s" % op.error_msg)
            if op.error_msg == "":
                op.set_status(OpStatus.STOPPED_BY_RECEIVER)
            else:
                op.set_status(OpStatus.FAILED)
        else:
            try:
                op.file_iterator.cancel()
            except AttributeError:
                # we may not have this yet if the transfer fails upon the initial response
                # (meaning we haven't returned the generator)
                pass
            logging.debug("Server: receiver received stop transfer by sender: %s" % op.error_msg)
            if op.error_msg == "":
                op.set_status(OpStatus.STOPPED_BY_SENDER)
            else:
                op.set_status(OpStatus.FAILED)

        return void



#!/usr/bin/python3

import gettext
import logging
from pathlib import Path

from gi.repository import GObject, GLib, Gio

import grpc

import transfers
import prefs
import misc
import util
import notifications
from util import OpStatus, OpCommand, TransferDirection, ReceiveError

_ = gettext.gettext

class CommonOp(GObject.Object):
    __gsignals__ = {
        "status-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
        "initial-setup-complete": (GObject.SignalFlags.RUN_LAST, None, ()),
        "op-command": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "progress-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
        "active": (GObject.SignalFlags.RUN_LAST, None, ()),
        "focus": (GObject.SignalFlags.RUN_LAST, None, ())
    }
    def __init__(self, direction, sender, uris=None):
        super(CommonOp, self).__init__()
        self.uris = uris
        self.sender = sender
        self.direction = direction
        self.status = OpStatus.INIT
        self.start_time = GLib.get_monotonic_time() # for sorting in the op list

        self.total_size = 0
        self.total_count = 0
        self.remaining_count = 0

        self.size_string = ""
        self.description = ""
        self.name_if_single = None
        self.mime_if_single = "application/octet-stream" # unknown
        self.gicon = Gio.content_type_get_symbolic_icon(self.mime_if_single)

        self.error_msg = ""

        self.progress_tracker = None

    def progress_report(self, report):
        self.current_progress_report = report
        report.progress_text = _("%(time_left)s (%(bytes_per_sec)s/s)") \
                                   % ({
                                         "time_left": util.format_time_span(report.time_left_sec),
                                         "bytes_per_sec": GLib.format_size(report.bytes_per_sec)
                                     })

        if report.progress == 1.0:
            self.status = OpStatus.FINISHED
            self.emit_status_changed()
            return

        self.emit("active")
        self.emit("progress-changed")

    def get_progress_text(self):
        try:
            return self.current_progress_report.progress_text
        except AttributeError:
            return ""

    def get_progress(self):
        try:
            return self.current_progress_report.progress
        except AttributeError:
            return 0

    def set_error(self, e=None):
        if e is None:
            self.error_msg = ""
            return

        if isinstance(e, GLib.Error):
            self.error_msg = e.message
        elif isinstance(e, grpc.RpcError):
            self.error_msg = e.details()
        elif isinstance(e, ReceiveError):
            self.error_msg = str(e)
        else:
            self.error_msg = str(e)

    @misc._idle
    def emit_initial_setup_complete(self):
        self.emit("initial-setup-complete")

    @misc._idle
    def emit_status_changed(self):
        self.emit("status-changed")

    def set_status(self, status):
        pass

    def focus(self):
        self.emit("focus")

class SendOp(CommonOp):
    def __init__(self, sender=None, receiver=None, receiver_name=None, uris=None):
        super(SendOp, self).__init__(TransferDirection.TO_REMOTE_MACHINE, sender, uris)
        self.receiver = receiver
        self.sender_name = GLib.get_real_name()
        self.receiver_name = receiver_name
        self.dbus_op = False
        self.resolved_files = []
        self.first_missing_file = None

        self.file_send_cancellable = None

        self.current_progress_report = None

        # These are the first-level base names (no path, just the filename) that we'll send to the server
        # to check for pre-existence.  We know that if these files/folders don't exist, none of their children
        # will.  This is a bit simple, but until we need more, it's fine.
        self.top_dir_basenames = []

    def set_status(self, status):
        self.status = status

        if status == OpStatus.FINISHED:
            notifications.TransferCompleteNotification(self, sender=True)
        elif status in (OpStatus.FAILED_UNRECOVERABLE, OpStatus.FAILED, OpStatus.FILE_NOT_FOUND):
            notifications.TransferFailedNotification(self, sender=True)
         # We only care if the other remote cancelled.  If we did it, we don't need a notification.
        elif status == OpStatus.STOPPED_BY_RECEIVER:
            notifications.TransferStoppedNotification(self, sender=True)

        self.emit_status_changed()

    def prepare_send_info(self):
        self.status = OpStatus.CALCULATING
        self.emit_status_changed()

        error = transfers.gather_file_info(self)

        self.update_ui_info(error)

    def update_ui_info(self, error):
        if error is None:
            self.size_string = GLib.format_size(self.total_size)
            logging.debug("Op: calculated %d files, with a size of %s" % (self.total_count, self.size_string))

            if self.total_count > 1:
                # Translators: Don't need to translate singular, we show the filename if there's only one
                self.description = gettext.ngettext("%d file-do-not-translate",
                                                    "%d files", self.total_count) % (self.total_count,)
                self.gicon = Gio.ThemedIcon.new("edit-copy-symbolic")
            else:
                self.description = self.resolved_files[0].basename
                self.gicon = Gio.content_type_get_symbolic_icon(self.mime_if_single)

            self.set_status(OpStatus.WAITING_PERMISSION)
        else:
            if isinstance(error, GLib.Error) and error.code == Gio.IOErrorEnum.NOT_FOUND:
                # self.status = OpStatus.FILE_NOT_FOUND
                self.description = ""
                self.error_msg = ""
                try:
                    self.first_missing_file = self.top_dir_basenames[0]
                except IndexError:
                    self.first_missing_file = None
                self.gicon = Gio.ThemedIcon.new("dialog-error-symbolic")
                self.set_status(OpStatus.FILE_NOT_FOUND)
            else:
                # self.status = OpStatus.FAILED_UNRECOVERABLE
                self.description = ""
                self.set_error(error)
                self.set_status(OpStatus.FAILED_UNRECOVERABLE)

        if self.dbus_op and self.status == OpStatus.WAITING_PERMISSION:
            notifications.WarpinatorSendNotification(self)

        self.emit_initial_setup_complete()
        self.emit_status_changed()

    # Widget handlers

    def cancel_transfer_request(self):
        self.emit("op-command", OpCommand.CANCEL_PERMISSION_BY_SENDER)

    def retry_transfer(self):
        self.emit("op-command", OpCommand.RETRY_TRANSFER)

    def pause_transfer(self):
        pass

    def stop_transfer(self):
        self.emit("op-command", OpCommand.STOP_TRANSFER_BY_SENDER)

    def remove_transfer(self):
        self.emit("op-command", OpCommand.REMOVE_TRANSFER)

# This represents a send or receive 'job', there would be potentially many of these.
class ReceiveOp(CommonOp):
    def __init__(self, sender):
        super(ReceiveOp, self).__init__(TransferDirection.FROM_REMOTE_MACHINE, sender)
        self.sender_name = self.sender
        self.receiver_name = GLib.get_real_name()

         # If there's insufficient disk space, always ask for permission
         # If we're overwriting, there's a preference to check whether we need to ask or not.
        self.have_space = False
        self.existing = False

        # This will be a <_Rendezvous object of in-flight RPC> if no compression is used, and a
        # SteamResponseWrapper (interceptors.py) if it is.
        self.file_iterator = None

        self.current_progress_report = None
        # These are the first-level base names (no path, just the filename) that we'll send to the server
        # to check for pre-existence.  We know that if these files/folders don't exist, none of their children
        # will.  This is a bit simple, but until we need more, it's fine.
        self.top_dir_basenames = []

    def set_status(self, status):
        self.status = status

        if status == OpStatus.FINISHED:
            notifications.TransferCompleteNotification(self, sender=False)
        elif status == OpStatus.FINISHED_WARNING:
            notifications.TransferCompleteNotification(self, sender=False, warn=True)
        elif status in (OpStatus.FAILED_UNRECOVERABLE, OpStatus.FAILED):
            notifications.TransferFailedNotification(self, sender=False)
         # We only care if the other remote cancelled.  If we did it, we don't need a notification.
        elif status == OpStatus.STOPPED_BY_SENDER:
            notifications.TransferStoppedNotification(self, sender=False)

        self.emit_status_changed()

    def prepare_receive_info(self):
        self.size_string = GLib.format_size(self.total_size)
        logging.debug("Op: details: %d files, with a size of %s" % (self.total_count, self.size_string))

        # Check that toplevels are valid, safe. This is done immediately to prevent some sort of runaway
        # free-space check.
        for top_dir in self.top_dir_basenames:
            try:
                util.test_resolved_path_safety(top_dir)
            except ReceiveError as e:
                self.set_error(e)
                self.status = OpStatus.FAILED_UNRECOVERABLE
                self.emit_initial_setup_complete()
                return

        self.have_space = util.free_space_monitor.have_enough_free(self.total_size, self.top_dir_basenames)
        self.existing = util.files_exist(self.top_dir_basenames)
        self.update_ui_info()

    def update_ui_info(self):
        if self.total_count > 1:
            # Translators: Don't need to translate singular, we show the filename if there's only one
            self.description = gettext.ngettext("%d file",
                                                "%d files", self.total_count) % (self.total_count,)
            self.gicon = Gio.ThemedIcon.new("edit-copy-symbolic")
        else:
            self.description = self.name_if_single
            self.gicon = Gio.content_type_get_symbolic_icon(self.mime_if_single)

        self.status = OpStatus.WAITING_PERMISSION

        notifications.NewOpUserNotification(self)
        self.emit_initial_setup_complete()

    # Widget handlers
    def accept_transfer(self):
        self.emit("op-command", OpCommand.START_TRANSFER)

    def decline_transfer_request(self):
        self.emit("op-command", OpCommand.CANCEL_PERMISSION_BY_RECEIVER)

    def stop_transfer(self):
        self.emit("op-command", OpCommand.STOP_TRANSFER_BY_RECEIVER)

    def remove_transfer(self):
        self.emit("op-command", OpCommand.REMOVE_TRANSFER)




#!/usr/bin/python3

import datetime
import stat
import os
from pathlib import Path
import secrets
import logging
import ipaddress

from cryptography import x509
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.x509.oid import NameOID
import nacl
from nacl import secret
import hashlib
import base64

from gi.repository import GLib, GObject

import util
import prefs

day = datetime.timedelta(1, 0, 0)
EXPIRE_TIME = 30 * day

singleton = None

def get_singleton():
    global singleton

    if singleton is None:
        singleton = AuthManager()

    return singleton

class AuthManager(GObject.Object):
    __gsignals__ = {
        'group-code-changed': (GObject.SignalFlags.RUN_LAST, None, ())
    }

    def __init__(self):
        GObject.Object.__init__(self)
        self.hostname = util.get_hostname()
        self.ip_info = None
        self.port = None

        self.private_key = None
        self.server_cert = None

        self.remote_certs = {}

        prefs.prefs_settings.connect("changed::group-code", self.notify_group_code_changed)

    def notify_group_code_changed(self, settings, key, data=None):
        self.emit("group-code-changed")

    def update(self, ip_info, port):
        self.ip_info = ip_info;
        self.port = port

        self._make_key_cert_pair()

    def get_server_creds(self):
        return (self.server_private_key, self.server_pub_key)

    def get_cached_cert(self, hostname, ip_info):
        try:
            return self.remote_certs["%s.%s" % (hostname, ip_info.ip4_address)]
        except KeyError:
            return None

    def process_remote_cert(self, hostname, ip_info, server_data):
        if server_data is None:
            return False
        decoded = base64.decodebytes(server_data)

        hasher = hashlib.sha256()
        hasher.update(bytes(prefs.get_group_code(), "utf-8"))
        key = hasher.digest()
        decoder = secret.SecretBox(key)

        try:
            cert = decoder.decrypt(decoded)
        except nacl.exceptions.CryptoError as e:
            logging.debug("Decryption failed for remote '%s': %s" % (hostname, str(e)))
            cert = None

        if cert:
            self.remote_certs["%s.%s" % (hostname, ip_info.ip4_address)] = cert
            return True
        else:
            return False

    def get_encoded_local_cert(self):
        hasher = hashlib.sha256()
        hasher.update(bytes(prefs.get_group_code(), "utf-8"))
        key = hasher.digest()

        encoder = secret.SecretBox(key)

        encrypted = encoder.encrypt(self.server_pub_key)
        encoded = base64.encodebytes(encrypted)
        return encoded

    def _make_key_cert_pair(self):
        logging.debug("Auth: Creating server credentials")

        private_key = rsa.generate_private_key(
            backend=crypto_default_backend(),
            public_exponent=65537,
            key_size=2048
        )

        public_key = private_key.public_key()

        builder = x509.CertificateBuilder()
        builder = builder.subject_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, self.hostname),
        ]))
        builder = builder.issuer_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, self.hostname),
        ]))
        builder = builder.not_valid_before(datetime.datetime.today() - day)
        builder = builder.not_valid_after(datetime.datetime.today() + EXPIRE_TIME)
        builder = builder.serial_number(x509.random_serial_number())
        builder = builder.public_key(public_key)

        alt_names = []

        if self.ip_info.ip4_address is not None:
            alt_names.append(x509.IPAddress(ipaddress.IPv4Address(self.ip_info.ip4_address)))

        builder = builder.add_extension(x509.SubjectAlternativeName(alt_names), critical=True)

        certificate = builder.sign(
            private_key=private_key, algorithm=hashes.SHA256(),
            backend=crypto_default_backend()
        )

        ser_private_key = private_key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.PKCS8,
            crypto_serialization.NoEncryption())

        ser_public_key = certificate.public_bytes(
            crypto_serialization.Encoding.PEM
        )

        self.server_pub_key = ser_public_key
        self.server_private_key = ser_private_key



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import warp_pb2 as warp__pb2


class WarpStub(object):
  """************ Important! ***************

  If you change anything here, you *must* run 'generate-protobuf' to update the
  generated stub files.

  Never change the existing members and member values of messages, only add new ones.

  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.CheckDuplexConnection = channel.unary_unary(
        '/Warp/CheckDuplexConnection',
        request_serializer=warp__pb2.LookupName.SerializeToString,
        response_deserializer=warp__pb2.HaveDuplex.FromString,
        )
    self.WaitingForDuplex = channel.unary_unary(
        '/Warp/WaitingForDuplex',
        request_serializer=warp__pb2.LookupName.SerializeToString,
        response_deserializer=warp__pb2.HaveDuplex.FromString,
        )
    self.GetRemoteMachineInfo = channel.unary_unary(
        '/Warp/GetRemoteMachineInfo',
        request_serializer=warp__pb2.LookupName.SerializeToString,
        response_deserializer=warp__pb2.RemoteMachineInfo.FromString,
        )
    self.GetRemoteMachineAvatar = channel.unary_stream(
        '/Warp/GetRemoteMachineAvatar',
        request_serializer=warp__pb2.LookupName.SerializeToString,
        response_deserializer=warp__pb2.RemoteMachineAvatar.FromString,
        )
    self.ProcessTransferOpRequest = channel.unary_unary(
        '/Warp/ProcessTransferOpRequest',
        request_serializer=warp__pb2.TransferOpRequest.SerializeToString,
        response_deserializer=warp__pb2.VoidType.FromString,
        )
    self.PauseTransferOp = channel.unary_unary(
        '/Warp/PauseTransferOp',
        request_serializer=warp__pb2.OpInfo.SerializeToString,
        response_deserializer=warp__pb2.VoidType.FromString,
        )
    self.StartTransfer = channel.unary_stream(
        '/Warp/StartTransfer',
        request_serializer=warp__pb2.OpInfo.SerializeToString,
        response_deserializer=warp__pb2.FileChunk.FromString,
        )
    self.CancelTransferOpRequest = channel.unary_unary(
        '/Warp/CancelTransferOpRequest',
        request_serializer=warp__pb2.OpInfo.SerializeToString,
        response_deserializer=warp__pb2.VoidType.FromString,
        )
    self.StopTransfer = channel.unary_unary(
        '/Warp/StopTransfer',
        request_serializer=warp__pb2.StopInfo.SerializeToString,
        response_deserializer=warp__pb2.VoidType.FromString,
        )
    self.Ping = channel.unary_unary(
        '/Warp/Ping',
        request_serializer=warp__pb2.LookupName.SerializeToString,
        response_deserializer=warp__pb2.VoidType.FromString,
        )


class WarpServicer(object):
  """************ Important! ***************

  If you change anything here, you *must* run 'generate-protobuf' to update the
  generated stub files.

  Never change the existing members and member values of messages, only add new ones.

  """

  def CheckDuplexConnection(self, request, context):
    """Sender methods
    api v1 duplex method (ping style)
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def WaitingForDuplex(self, request, context):
    """api v2 duplex method (block/future)
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetRemoteMachineInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetRemoteMachineAvatar(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ProcessTransferOpRequest(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PauseTransferOp(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def StartTransfer(self, request, context):
    """Receiver methods
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CancelTransferOpRequest(self, request, context):
    """Both
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def StopTransfer(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Ping(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_WarpServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'CheckDuplexConnection': grpc.unary_unary_rpc_method_handler(
          servicer.CheckDuplexConnection,
          request_deserializer=warp__pb2.LookupName.FromString,
          response_serializer=warp__pb2.HaveDuplex.SerializeToString,
      ),
      'WaitingForDuplex': grpc.unary_unary_rpc_method_handler(
          servicer.WaitingForDuplex,
          request_deserializer=warp__pb2.LookupName.FromString,
          response_serializer=warp__pb2.HaveDuplex.SerializeToString,
      ),
      'GetRemoteMachineInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetRemoteMachineInfo,
          request_deserializer=warp__pb2.LookupName.FromString,
          response_serializer=warp__pb2.RemoteMachineInfo.SerializeToString,
      ),
      'GetRemoteMachineAvatar': grpc.unary_stream_rpc_method_handler(
          servicer.GetRemoteMachineAvatar,
          request_deserializer=warp__pb2.LookupName.FromString,
          response_serializer=warp__pb2.RemoteMachineAvatar.SerializeToString,
      ),
      'ProcessTransferOpRequest': grpc.unary_unary_rpc_method_handler(
          servicer.ProcessTransferOpRequest,
          request_deserializer=warp__pb2.TransferOpRequest.FromString,
          response_serializer=warp__pb2.VoidType.SerializeToString,
      ),
      'PauseTransferOp': grpc.unary_unary_rpc_method_handler(
          servicer.PauseTransferOp,
          request_deserializer=warp__pb2.OpInfo.FromString,
          response_serializer=warp__pb2.VoidType.SerializeToString,
      ),
      'StartTransfer': grpc.unary_stream_rpc_method_handler(
          servicer.StartTransfer,
          request_deserializer=warp__pb2.OpInfo.FromString,
          response_serializer=warp__pb2.FileChunk.SerializeToString,
      ),
      'CancelTransferOpRequest': grpc.unary_unary_rpc_method_handler(
          servicer.CancelTransferOpRequest,
          request_deserializer=warp__pb2.OpInfo.FromString,
          response_serializer=warp__pb2.VoidType.SerializeToString,
      ),
      'StopTransfer': grpc.unary_unary_rpc_method_handler(
          servicer.StopTransfer,
          request_deserializer=warp__pb2.StopInfo.FromString,
          response_serializer=warp__pb2.VoidType.SerializeToString,
      ),
      'Ping': grpc.unary_unary_rpc_method_handler(
          servicer.Ping,
          request_deserializer=warp__pb2.LookupName.FromString,
          response_serializer=warp__pb2.VoidType.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'Warp', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class WarpRegistrationStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.RequestCertificate = channel.unary_unary(
        '/WarpRegistration/RequestCertificate',
        request_serializer=warp__pb2.RegRequest.SerializeToString,
        response_deserializer=warp__pb2.RegResponse.FromString,
        )


class WarpRegistrationServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def RequestCertificate(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_WarpRegistrationServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'RequestCertificate': grpc.unary_unary_rpc_method_handler(
          servicer.RequestCertificate,
          request_deserializer=warp__pb2.RegRequest.FromString,
          response_serializer=warp__pb2.RegResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'WarpRegistration', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



"""Landlock constants and syscalls."""
import ctypes
import enum
import errno
import functools
import os
import platform
from typing import Callable, Optional, Tuple, TypeVar

import _ctypes

from landlock import SyscallError

CREATE_RULESET_VERSION = 1 << 0

SYSCALL_CREATE_RULESET = 444
SYSCALL_ADD_RULE = 445
SYSCALL_RESTRICT_SELF = 446
PR_SET_NO_NEW_PRIVS = 38

T = TypeVar("T")


class FSAccess(enum.IntFlag):
    """Flags representing types of file system actions.

    These flags enable one to restrict a sandboxed process
    to a set of actions on files and directories.

    Files or directories opened before the sandboxing
    are not subject to these restrictions.
    """

    # A file can only receive these access rights:
    EXECUTE = 1 << 0
    """Execute a file."""
    WRITE_FILE = 1 << 1
    """Open a file with write access."""
    READ_FILE = 1 << 2
    """Open a file with read access."""

    # A directory can receive access rights related to files or directories.
    # The following access right is applied to the directory itself,
    # and the directories beneath it:
    READ_DIR = 1 << 3
    """Open a directory or list its content."""

    # However, the following access rights only apply to the content of a directory,
    # not the directory itself:
    REMOVE_DIR = 1 << 4
    """Remove an empty directory or rename one."""
    REMOVE_FILE = 1 << 5
    """Unlink (or rename) a file."""
    MAKE_CHAR = 1 << 6
    """Create (or rename or link) a character device."""
    MAKE_DIR = 1 << 7
    """Create (or rename) a directory."""
    MAKE_REG = 1 << 8
    """Create (or rename or link) a regular file."""
    MAKE_SOCK = 1 << 9
    """Create (or rename or link) a UNIX domain socket."""
    MAKE_FIFO = 1 << 10
    """Create (or rename or link) a named pipe."""
    MAKE_BLOCK = 1 << 11
    """Create (or rename or link) a block device."""
    MAKE_SYM = 1 << 12
    """Create (or rename or link) a symbolic link."""
    REFER = 1 << 13
    """Link or rename a file from or to a different directory.

    I.e. reparent a file hierarchy.

    Only available if the ABI version >= 2.
    """

    @classmethod
    def all(cls):
        return cls.all_file() | cls.all_dir()

    @classmethod
    def all_file(cls):
        return cls.EXECUTE | cls.WRITE_FILE | cls.READ_FILE

    @classmethod
    def all_dir(cls):
        flags = (
            cls.READ_DIR
            | cls.REMOVE_DIR
            | cls.REMOVE_FILE
            | cls.MAKE_CHAR
            | cls.MAKE_DIR
            | cls.MAKE_REG
            | cls.MAKE_SOCK
            | cls.MAKE_FIFO
            | cls.MAKE_BLOCK
            | cls.MAKE_SYM
        )
        # REFER only available in version 2
        if landlock_abi_version() >= 2:
            flags |= cls.REFER
        return flags


class RulesetAttr(ctypes.Structure):
    _fields_ = [("handled_access_fs", ctypes.c_uint64)]


class PathBeneathAttr(ctypes.Structure):
    _fields_ = [
        ("allowed_access", ctypes.c_uint64),
        ("parent_fd", ctypes.c_int32),
    ]


def find_generic_reason_from_platform() -> Optional[str]:
    # check OS
    system = platform.system()
    if system != "Linux":
        return (
            f"Landlock is a only available on Linux,"
            f" it looks like you're running {system}"
        )

    # check kernel version
    kernel_version = platform.release()
    kernel_version_tuple = tuple(map(int, kernel_version.split("-")[0].split(".")))
    if kernel_version_tuple < (5, 13):
        return (
            f"Landlock is only available in kernel 5.13 or newer,"
            f" it looks like you're running {kernel_version}"
        )

    return None


def find_generic_reason_from_errno(err: int) -> Optional[str]:
    errno_to_reason = {
        errno.ENOSYS: "Cannot find Landlock syscalls - perhaps the kernel was not built"
        " with 'CONFIG_SECURITY_LANDLOCK=y'",
        errno.EOPNOTSUPP: "Landlock has been disabled at boot time."
        " The CONFIG_LSM configuration item should list 'landlock',"
        " or, 'lsm=landlock' needs to be added the kernel's"
        " command-line arguments (usually via your bootloader).",
    }
    return errno_to_reason.get(err)


def syscall_errcheck(
    result: T, func: _ctypes.CFuncPtr, arguments: Tuple, reason: Optional[str] = None
) -> T:
    err = ctypes.get_errno()

    # return the result if there was no error
    if err == 0:
        return result

    # otherwise raise a SyscallError exception
    name = getattr(func, "name", func.__name__)
    reason = (
        find_generic_reason_from_platform()
        or find_generic_reason_from_errno(err)
        or reason
    )
    raise SyscallError(
        err,
        os.strerror(err),
        f"Error calling {name}: {func.__name__}{arguments} = {result}",
        reason=reason,
    )


def create_ruleset_errcheck(result: T, func: _ctypes.CFuncPtr, arguments: Tuple) -> T:
    errno_to_reason = {
        errno.EINVAL: "Unknown 'flags', or unknown access, or too small 'size'",
        errno.E2BIG: "'attr' or 'size' inconsistencies",
        errno.EFAULT: "'attr' or 'size' inconsistencies",
        errno.ENOMSG: "Empty 'landlock_ruleset_attr.handled_access_fs'",
    }
    return syscall_errcheck(
        result, func, arguments, errno_to_reason.get(ctypes.get_errno())
    )


def add_rule_errcheck(result: T, func: _ctypes.CFuncPtr, arguments: Tuple) -> T:
    errno_to_reason = {
        errno.EINVAL: "'flags' is not 0, or inconsistent access in the rule"
        " (i.e. 'landlock_path_beneath_attr.allowed_access'"
        " is not a subset of the ruleset handled accesses)",
        errno.ENOMSG: "Empty accesses"
        " (e.g. 'landlock_path_beneath_attr.allowed_access')",
        errno.EBADF: "'ruleset_fd' is not a file descriptor for the current thread,"
        " or a member of 'rule_attr' is not a file descriptor as expected",
        errno.EBADFD: "'ruleset_fd' is not a ruleset file descriptor,"
        " or a member of 'rule_attr' is not the expected file descriptor type",
        errno.EPERM: "'ruleset_fd' has no write access to the underlying ruleset",
        errno.EFAULT: "'rule_attr' inconsistency",
    }
    return syscall_errcheck(
        result, func, arguments, errno_to_reason.get(ctypes.get_errno())
    )


def restrict_self_errcheck(result: T, func: _ctypes.CFuncPtr, arguments: Tuple) -> T:
    errno_to_reason = {
        errno.EINVAL: "'flags' is not 0",
        errno.EBADF: "'ruleset_fd' is not a file descriptor for the current thread",
        errno.EBADFD: "'ruleset_fd' is not a ruleset file descriptor",
        errno.EPERM: "'ruleset_fd' has no read access to the underlying ruleset,"
        " or the current thread is not running with no_new_privs,"
        " or it doesn’t have 'CAP_SYS_ADMIN' in its namespace",
        errno.E2BIG: "The maximum number of stacked rulesets (16)"
        " has been reached for the current thread",
    }
    return syscall_errcheck(
        result, func, arguments, errno_to_reason.get(ctypes.get_errno())
    )


@functools.lru_cache(1)
def get_libc() -> ctypes.CDLL:
    try:
        return ctypes.CDLL(None, use_errno=True)
    except TypeError as e:
        if platform.system() == "Windows":
            # on Windows we get a TypeError using name=None
            raise SyscallError(reason=find_generic_reason_from_platform()) from e
        raise


@functools.lru_cache(1)
def get_create_ruleset() -> Callable:
    libc = get_libc()

    create_ruleset = functools.partial(libc["syscall"], SYSCALL_CREATE_RULESET)
    create_ruleset.func.name = "landlock_create_ruleset"
    create_ruleset.func.argtypes = (
        ctypes.c_long,
        ctypes.POINTER(RulesetAttr),
        ctypes.c_size_t,
        ctypes.c_uint32,
    )
    create_ruleset.func.errcheck = create_ruleset_errcheck
    create_ruleset.func.restype = ctypes.c_long

    return create_ruleset


@functools.lru_cache(1)
def get_add_rule() -> Callable:
    libc = get_libc()

    add_rule = functools.partial(libc["syscall"], SYSCALL_ADD_RULE)
    add_rule.func.name = "landlock_add_rule"
    add_rule.func.argtypes = (
        ctypes.c_long,
        ctypes.c_int,
        ctypes.c_uint,
        ctypes.POINTER(PathBeneathAttr),
        ctypes.c_uint32,
    )
    add_rule.func.errcheck = add_rule_errcheck
    add_rule.func.restype = ctypes.c_long

    return add_rule


@functools.lru_cache(1)
def get_restrict_self() -> Callable:
    libc = get_libc()

    restrict_self = functools.partial(libc["syscall"], SYSCALL_RESTRICT_SELF)
    restrict_self.func.name = "landlock_restrict_self"
    restrict_self.func.argtypes = (
        ctypes.c_long,
        ctypes.c_int,
        ctypes.c_uint32,
    )
    restrict_self.func.errcheck = restrict_self_errcheck
    restrict_self.func.restype = ctypes.c_long

    return restrict_self


@functools.lru_cache(1)
def get_prctl() -> Callable:
    # https://man7.org/linux/man-pages/man2/prctl.2.html
    libc = get_libc()

    prctl = libc.prctl
    prctl.name = "prctl"
    prctl.argtypes = (
        ctypes.c_int,
        ctypes.c_ulong,
        ctypes.c_ulong,
        ctypes.c_ulong,
        ctypes.c_ulong,
    )
    prctl.errcheck = syscall_errcheck
    prctl.restype = ctypes.c_int

    return libc.prctl


@functools.lru_cache(1)
def landlock_abi_version() -> int:
    return get_create_ruleset()(None, 0, CREATE_RULESET_VERSION)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: warp.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='warp.proto',
  package='',
  syntax='proto3',
  serialized_pb=_b('\n\nwarp.proto\"<\n\x11RemoteMachineInfo\x12\x14\n\x0c\x64isplay_name\x18\x01 \x01(\t\x12\x11\n\tuser_name\x18\x02 \x01(\t\"+\n\x13RemoteMachineAvatar\x12\x14\n\x0c\x61vatar_chunk\x18\x01 \x01(\x0c\"/\n\nLookupName\x12\n\n\x02id\x18\x01 \x01(\t\x12\x15\n\rreadable_name\x18\x02 \x01(\t\"\x1e\n\nHaveDuplex\x12\x10\n\x08response\x18\x02 \x01(\x08\"\x19\n\x08VoidType\x12\r\n\x05\x64ummy\x18\x01 \x01(\x05\"Z\n\x06OpInfo\x12\r\n\x05ident\x18\x01 \x01(\t\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\x12\x15\n\rreadable_name\x18\x03 \x01(\t\x12\x17\n\x0fuse_compression\x18\x04 \x01(\x08\"0\n\x08StopInfo\x12\x15\n\x04info\x18\x01 \x01(\x0b\x32\x07.OpInfo\x12\r\n\x05\x65rror\x18\x02 \x01(\x08\"\xd0\x01\n\x11TransferOpRequest\x12\x15\n\x04info\x18\x01 \x01(\x0b\x32\x07.OpInfo\x12\x13\n\x0bsender_name\x18\x02 \x01(\t\x12\x15\n\rreceiver_name\x18\x03 \x01(\t\x12\x10\n\x08receiver\x18\x04 \x01(\t\x12\x0c\n\x04size\x18\x05 \x01(\x04\x12\r\n\x05\x63ount\x18\x06 \x01(\x04\x12\x16\n\x0ename_if_single\x18\x07 \x01(\t\x12\x16\n\x0emime_if_single\x18\x08 \x01(\t\x12\x19\n\x11top_dir_basenames\x18\t \x03(\t\"\x88\x01\n\tFileChunk\x12\x15\n\rrelative_path\x18\x01 \x01(\t\x12\x11\n\tfile_type\x18\x02 \x01(\x05\x12\x16\n\x0esymlink_target\x18\x03 \x01(\t\x12\r\n\x05\x63hunk\x18\x04 \x01(\x0c\x12\x11\n\tfile_mode\x18\x05 \x01(\r\x12\x17\n\x04time\x18\x06 \x01(\x0b\x32\t.FileTime\"-\n\x08\x46ileTime\x12\r\n\x05mtime\x18\x01 \x01(\x04\x12\x12\n\nmtime_usec\x18\x02 \x01(\r\"*\n\nRegRequest\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x10\n\x08hostname\x18\x02 \x01(\t\"\"\n\x0bRegResponse\x12\x13\n\x0blocked_cert\x18\x01 \x01(\t2\xf2\x03\n\x04Warp\x12\x33\n\x15\x43heckDuplexConnection\x12\x0b.LookupName\x1a\x0b.HaveDuplex\"\x00\x12.\n\x10WaitingForDuplex\x12\x0b.LookupName\x1a\x0b.HaveDuplex\"\x00\x12\x39\n\x14GetRemoteMachineInfo\x12\x0b.LookupName\x1a\x12.RemoteMachineInfo\"\x00\x12?\n\x16GetRemoteMachineAvatar\x12\x0b.LookupName\x1a\x14.RemoteMachineAvatar\"\x00\x30\x01\x12;\n\x18ProcessTransferOpRequest\x12\x12.TransferOpRequest\x1a\t.VoidType\"\x00\x12\'\n\x0fPauseTransferOp\x12\x07.OpInfo\x1a\t.VoidType\"\x00\x12(\n\rStartTransfer\x12\x07.OpInfo\x1a\n.FileChunk\"\x00\x30\x01\x12/\n\x17\x43\x61ncelTransferOpRequest\x12\x07.OpInfo\x1a\t.VoidType\"\x00\x12&\n\x0cStopTransfer\x12\t.StopInfo\x1a\t.VoidType\"\x00\x12 \n\x04Ping\x12\x0b.LookupName\x1a\t.VoidType\"\x00\x32\x45\n\x10WarpRegistration\x12\x31\n\x12RequestCertificate\x12\x0b.RegRequest\x1a\x0c.RegResponse\"\x00\x62\x06proto3')
)




_REMOTEMACHINEINFO = _descriptor.Descriptor(
  name='RemoteMachineInfo',
  full_name='RemoteMachineInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_name', full_name='RemoteMachineInfo.display_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_name', full_name='RemoteMachineInfo.user_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14,
  serialized_end=74,
)


_REMOTEMACHINEAVATAR = _descriptor.Descriptor(
  name='RemoteMachineAvatar',
  full_name='RemoteMachineAvatar',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='avatar_chunk', full_name='RemoteMachineAvatar.avatar_chunk', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=76,
  serialized_end=119,
)


_LOOKUPNAME = _descriptor.Descriptor(
  name='LookupName',
  full_name='LookupName',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='LookupName.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='readable_name', full_name='LookupName.readable_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=121,
  serialized_end=168,
)


_HAVEDUPLEX = _descriptor.Descriptor(
  name='HaveDuplex',
  full_name='HaveDuplex',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='response', full_name='HaveDuplex.response', index=0,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=170,
  serialized_end=200,
)


_VOIDTYPE = _descriptor.Descriptor(
  name='VoidType',
  full_name='VoidType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dummy', full_name='VoidType.dummy', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=202,
  serialized_end=227,
)


_OPINFO = _descriptor.Descriptor(
  name='OpInfo',
  full_name='OpInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ident', full_name='OpInfo.ident', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='OpInfo.timestamp', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='readable_name', full_name='OpInfo.readable_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='use_compression', full_name='OpInfo.use_compression', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=229,
  serialized_end=319,
)


_STOPINFO = _descriptor.Descriptor(
  name='StopInfo',
  full_name='StopInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='info', full_name='StopInfo.info', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='StopInfo.error', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=321,
  serialized_end=369,
)


_TRANSFEROPREQUEST = _descriptor.Descriptor(
  name='TransferOpRequest',
  full_name='TransferOpRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='info', full_name='TransferOpRequest.info', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sender_name', full_name='TransferOpRequest.sender_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receiver_name', full_name='TransferOpRequest.receiver_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receiver', full_name='TransferOpRequest.receiver', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='TransferOpRequest.size', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='count', full_name='TransferOpRequest.count', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name_if_single', full_name='TransferOpRequest.name_if_single', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mime_if_single', full_name='TransferOpRequest.mime_if_single', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='top_dir_basenames', full_name='TransferOpRequest.top_dir_basenames', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=372,
  serialized_end=580,
)


_FILECHUNK = _descriptor.Descriptor(
  name='FileChunk',
  full_name='FileChunk',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='relative_path', full_name='FileChunk.relative_path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_type', full_name='FileChunk.file_type', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symlink_target', full_name='FileChunk.symlink_target', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chunk', full_name='FileChunk.chunk', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file_mode', full_name='FileChunk.file_mode', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='time', full_name='FileChunk.time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=583,
  serialized_end=719,
)


_FILETIME = _descriptor.Descriptor(
  name='FileTime',
  full_name='FileTime',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mtime', full_name='FileTime.mtime', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mtime_usec', full_name='FileTime.mtime_usec', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=721,
  serialized_end=766,
)


_REGREQUEST = _descriptor.Descriptor(
  name='RegRequest',
  full_name='RegRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ip', full_name='RegRequest.ip', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hostname', full_name='RegRequest.hostname', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=768,
  serialized_end=810,
)


_REGRESPONSE = _descriptor.Descriptor(
  name='RegResponse',
  full_name='RegResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='locked_cert', full_name='RegResponse.locked_cert', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=812,
  serialized_end=846,
)

_STOPINFO.fields_by_name['info'].message_type = _OPINFO
_TRANSFEROPREQUEST.fields_by_name['info'].message_type = _OPINFO
_FILECHUNK.fields_by_name['time'].message_type = _FILETIME
DESCRIPTOR.message_types_by_name['RemoteMachineInfo'] = _REMOTEMACHINEINFO
DESCRIPTOR.message_types_by_name['RemoteMachineAvatar'] = _REMOTEMACHINEAVATAR
DESCRIPTOR.message_types_by_name['LookupName'] = _LOOKUPNAME
DESCRIPTOR.message_types_by_name['HaveDuplex'] = _HAVEDUPLEX
DESCRIPTOR.message_types_by_name['VoidType'] = _VOIDTYPE
DESCRIPTOR.message_types_by_name['OpInfo'] = _OPINFO
DESCRIPTOR.message_types_by_name['StopInfo'] = _STOPINFO
DESCRIPTOR.message_types_by_name['TransferOpRequest'] = _TRANSFEROPREQUEST
DESCRIPTOR.message_types_by_name['FileChunk'] = _FILECHUNK
DESCRIPTOR.message_types_by_name['FileTime'] = _FILETIME
DESCRIPTOR.message_types_by_name['RegRequest'] = _REGREQUEST
DESCRIPTOR.message_types_by_name['RegResponse'] = _REGRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RemoteMachineInfo = _reflection.GeneratedProtocolMessageType('RemoteMachineInfo', (_message.Message,), dict(
  DESCRIPTOR = _REMOTEMACHINEINFO,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:RemoteMachineInfo)
  ))
_sym_db.RegisterMessage(RemoteMachineInfo)

RemoteMachineAvatar = _reflection.GeneratedProtocolMessageType('RemoteMachineAvatar', (_message.Message,), dict(
  DESCRIPTOR = _REMOTEMACHINEAVATAR,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:RemoteMachineAvatar)
  ))
_sym_db.RegisterMessage(RemoteMachineAvatar)

LookupName = _reflection.GeneratedProtocolMessageType('LookupName', (_message.Message,), dict(
  DESCRIPTOR = _LOOKUPNAME,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:LookupName)
  ))
_sym_db.RegisterMessage(LookupName)

HaveDuplex = _reflection.GeneratedProtocolMessageType('HaveDuplex', (_message.Message,), dict(
  DESCRIPTOR = _HAVEDUPLEX,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:HaveDuplex)
  ))
_sym_db.RegisterMessage(HaveDuplex)

VoidType = _reflection.GeneratedProtocolMessageType('VoidType', (_message.Message,), dict(
  DESCRIPTOR = _VOIDTYPE,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:VoidType)
  ))
_sym_db.RegisterMessage(VoidType)

OpInfo = _reflection.GeneratedProtocolMessageType('OpInfo', (_message.Message,), dict(
  DESCRIPTOR = _OPINFO,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:OpInfo)
  ))
_sym_db.RegisterMessage(OpInfo)

StopInfo = _reflection.GeneratedProtocolMessageType('StopInfo', (_message.Message,), dict(
  DESCRIPTOR = _STOPINFO,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:StopInfo)
  ))
_sym_db.RegisterMessage(StopInfo)

TransferOpRequest = _reflection.GeneratedProtocolMessageType('TransferOpRequest', (_message.Message,), dict(
  DESCRIPTOR = _TRANSFEROPREQUEST,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:TransferOpRequest)
  ))
_sym_db.RegisterMessage(TransferOpRequest)

FileChunk = _reflection.GeneratedProtocolMessageType('FileChunk', (_message.Message,), dict(
  DESCRIPTOR = _FILECHUNK,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:FileChunk)
  ))
_sym_db.RegisterMessage(FileChunk)

FileTime = _reflection.GeneratedProtocolMessageType('FileTime', (_message.Message,), dict(
  DESCRIPTOR = _FILETIME,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:FileTime)
  ))
_sym_db.RegisterMessage(FileTime)

RegRequest = _reflection.GeneratedProtocolMessageType('RegRequest', (_message.Message,), dict(
  DESCRIPTOR = _REGREQUEST,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:RegRequest)
  ))
_sym_db.RegisterMessage(RegRequest)

RegResponse = _reflection.GeneratedProtocolMessageType('RegResponse', (_message.Message,), dict(
  DESCRIPTOR = _REGRESPONSE,
  __module__ = 'warp_pb2'
  # @@protoc_insertion_point(class_scope:RegResponse)
  ))
_sym_db.RegisterMessage(RegResponse)



_WARP = _descriptor.ServiceDescriptor(
  name='Warp',
  full_name='Warp',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=849,
  serialized_end=1347,
  methods=[
  _descriptor.MethodDescriptor(
    name='CheckDuplexConnection',
    full_name='Warp.CheckDuplexConnection',
    index=0,
    containing_service=None,
    input_type=_LOOKUPNAME,
    output_type=_HAVEDUPLEX,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='WaitingForDuplex',
    full_name='Warp.WaitingForDuplex',
    index=1,
    containing_service=None,
    input_type=_LOOKUPNAME,
    output_type=_HAVEDUPLEX,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetRemoteMachineInfo',
    full_name='Warp.GetRemoteMachineInfo',
    index=2,
    containing_service=None,
    input_type=_LOOKUPNAME,
    output_type=_REMOTEMACHINEINFO,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetRemoteMachineAvatar',
    full_name='Warp.GetRemoteMachineAvatar',
    index=3,
    containing_service=None,
    input_type=_LOOKUPNAME,
    output_type=_REMOTEMACHINEAVATAR,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='ProcessTransferOpRequest',
    full_name='Warp.ProcessTransferOpRequest',
    index=4,
    containing_service=None,
    input_type=_TRANSFEROPREQUEST,
    output_type=_VOIDTYPE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PauseTransferOp',
    full_name='Warp.PauseTransferOp',
    index=5,
    containing_service=None,
    input_type=_OPINFO,
    output_type=_VOIDTYPE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='StartTransfer',
    full_name='Warp.StartTransfer',
    index=6,
    containing_service=None,
    input_type=_OPINFO,
    output_type=_FILECHUNK,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='CancelTransferOpRequest',
    full_name='Warp.CancelTransferOpRequest',
    index=7,
    containing_service=None,
    input_type=_OPINFO,
    output_type=_VOIDTYPE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='StopTransfer',
    full_name='Warp.StopTransfer',
    index=8,
    containing_service=None,
    input_type=_STOPINFO,
    output_type=_VOIDTYPE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='Ping',
    full_name='Warp.Ping',
    index=9,
    containing_service=None,
    input_type=_LOOKUPNAME,
    output_type=_VOIDTYPE,
    options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_WARP)

DESCRIPTOR.services_by_name['Warp'] = _WARP


_WARPREGISTRATION = _descriptor.ServiceDescriptor(
  name='WarpRegistration',
  full_name='WarpRegistration',
  file=DESCRIPTOR,
  index=1,
  options=None,
  serialized_start=1349,
  serialized_end=1418,
  methods=[
  _descriptor.MethodDescriptor(
    name='RequestCertificate',
    full_name='WarpRegistration.RequestCertificate',
    index=0,
    containing_service=None,
    input_type=_REGREQUEST,
    output_type=_REGRESPONSE,
    options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_WARPREGISTRATION)

DESCRIPTOR.services_by_name['WarpRegistration'] = _WARPREGISTRATION

# @@protoc_insertion_point(module_scope)
